from openai.types.chat import ChatCompletionMessageParam


class ServiceUnavailableException(Exception):
    def __init__(self, message="Сервис недоступен. Возможно, требуется включить VPN или сервис не запущен."):
        super().__init__(message)


class UnsupportedModelException(Exception):
    def __init__(
        self, model_id, model_names, message="Модель {model_id} недоступна. Список доступных моделей - {model_names}"
    ):
        super().__init__(message.format(model_id=model_id, model_names=model_names))


EndpointLLMPayload = list[ChatCompletionMessageParam] | dict



from mmar_mapi.api import LLMRequest, LLMResponseExt

from mmar_llm.llm_endpoint import LLMEndpoint


class DummyEndpoint(LLMEndpoint):
    """for tests"""

    def get_response_ext(self, *, request: LLMRequest) -> LLMResponseExt:
        return f"Request: {request}"

    def get_embedding(self, *, prompt: str) -> list[float]:
        return [1, 2, 3]



import base64

from openai import DefaultHttpxClient, OpenAI

from mmar_llm.llm_endpoint import LLMEndpoint
from mmar_llm.utils import dump_messages
from mmar_mapi.api import LLMPayload, LLMRequest, LLMResponseExt

OPENROUTER_BASE_URL = "https://openrouter.ai/api/v1"


class OpenRouterEndpoint(LLMEndpoint):
    def __init__(
        self,
        model_id: str,
        api_key: str,
        base_url=OPENROUTER_BASE_URL,
        emb_dim: int = 1024,
        providers: list[str] = [],
        verify: bool = True,
    ) -> None:
        self.base_url = base_url
        self._model = OpenAI(base_url=base_url, api_key=api_key, http_client=DefaultHttpxClient(verify=verify))
        self.model_id: str = model_id
        self.extra_body: dict[str, dict[str, list[str]]] = {"provider": {"order": providers}}
        self._dim: int = emb_dim

    def __call__(self) -> OpenAI:
        return self._model

    def get_response_ext(self, *, request: LLMRequest) -> LLMResponseExt:
        payload = LLMPayload.parse(request)
        messages_json = dump_messages(payload)

        completions = self._model.chat.completions
        response_openai = completions.create(model=self.model_id, messages=messages_json, extra_body=self.extra_body)
        text = response_openai.choices[0].message.content or ""
        return LLMResponseExt(text=text)

    def _create_image_payload(
        self, system_prompt: str, user_prompt: str, image_encoded: str, mimetype: str = "image/jpeg"
    ):
        image_content = [
            {"type": "text", "text": user_prompt},
            {"type": "image_url", "image_url": {"url": f"data:{mimetype};base64,{image_encoded}"}},
        ]
        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": image_content},
        ]

    # todo fix, call via #get_response_ext
    def get_image_response(self, bytesimage: bytes, sentences: str, mimetype: str = "image/jpeg") -> str:
        encoded_image = base64.b64encode(bytesimage).decode("utf-8")
        payload = self._create_image_payload(
            system_prompt="", user_prompt=sentences, image_encoded=encoded_image, mimetype=mimetype
        )
        completions = self._model.chat.completions
        response_openai = completions.create(model=self.model_id, messages=payload, extra_body=self.extra_body)
        text = response_openai.choices[0].message.content or ""
        return text

    def get_embedding(self, *, prompt: str) -> list[float]:
        return self._model.embeddings.create(model=self.model_id, input=[prompt]).data[0].embedding

    def __repr__(self):
        class_name = type(self).__name__
        url_info = f", url={self.base_url}" if self.base_url != OPENROUTER_BASE_URL else ""
        return f"{class_name}(model_id={self.model_id}{url_info})"

    def __str__(self):
        return repr(self)



import mimetypes
import time
from functools import cache
from pathlib import Path

from loguru import logger
from mmar_mapi import FileStorage
from mmar_mapi.services import (
    LCP,
    RESPONSE_EMPTY,
    LLMCallProps,
    LLMHubAPI,
    LLMHubMetadata,
    LLMPayload,
    LLMRequest,
    LLMResponseExt,
    ResourceId,
)
from mmar_utils import pretty_line, retry_on_cond
from requests.exceptions import ConnectTimeout

from mmar_llm.gigachat_endpoint import GigaChatEndpoint
from mmar_llm.llm_endpoint import LLMEndpoint
from mmar_llm.llm_hub_config import LLMConfig, LLMHubConfig
from mmar_llm.models import ServiceUnavailableException
from mmar_llm.openrouter_endpoint import OpenRouterEndpoint
from mmar_llm.utils import load_dynamically

ENDPOINTS_CAPABILITIES: dict[str, type | tuple[type]] = {
    "image": OpenRouterEndpoint,
    "file": GigaChatEndpoint,
}
IMAGE_MIME_TYPES = {
    "jpg": "image/jpeg",
    "png": "image/png",
}
NA = "NOT AVAILABLE"


def _parse_prompt_for_image(payload: LLMPayload) -> str:
    messages = payload.messages
    m_len = len(messages)
    if m_len == 0:
        return ""
    if m_len > 1:
        logger.warning(f"One message expected, but passed {m_len}")
    msg = messages[-1]
    return msg.content


class LLMHub(LLMHubAPI):
    def __init__(self, config: LLMHubConfig):
        config_llm = config.llm
        self.config_llm: LLMConfig = config_llm

        self.wait_seconds = config_llm.wait_seconds_on_llm_retry
        self.endpoint_keys = [ep.key for ep in config_llm.endpoints]

        if config_llm.warmup:
            eks_loaded = {ek for ek in self.endpoint_keys if self._get_endpoint(ek) is not None}
            endpoints_pretty = ", ".join(eks_loaded)
            logger.info(f"Ready endpoints: {endpoints_pretty}")
            na_endpoints_pretty = ", ".join(ek for ek in self.endpoint_keys if ek not in eks_loaded)
            logger.warning(f"Not available endpoints: {na_endpoints_pretty}")
        else:
            endpoints_pretty = ", ".join(self.endpoint_keys)
            logger.info(f"Endpoints: {endpoints_pretty}")

        self.default_ek = config_llm.default_endpoint_key
        self.default_image_ek = config_llm.default_image_endpoint_key
        self.default_file_ek = config_llm.default_file_endpoint_key
        self.file_storage = FileStorage.create(config.files_dir)
        self.validate_endpoints = getattr(config, "validate_endpoints", True)

    @cache
    def _get_endpoint(self, ek: str, capability: str | None = None) -> LLMEndpoint | None:
        try:
            ep_configs = [epc for epc in self.config_llm.endpoints if epc.key == ek]
            if len(ep_configs) == 0:
                logger.warning(f"Not found endpoint with key={ek}")
                return None
            if len(ep_configs) > 1:
                logger.warning(f"Found many entrypoint configs: {ep_configs}")
            ep_config = ep_configs[0]
            ep_class = load_dynamically(ep_config.class_path)
            ep = ep_class(**ep_config.args)
            if self.validate_endpoints:
                if not isinstance(ep, LLMEndpoint):
                    raise ValueError(f"Expected LLMEndpoint, but found {type(ep)}: {ep}")
                # todo validate methods
            if ep is None:
                logger.warning(f"Not found endpoint with key={ek}")
                return None
            if not capability:
                return ep
            cls = ENDPOINTS_CAPABILITIES[capability]
            if isinstance(ep, cls):
                return ep
            logger.warning(f"Expected endpoint {cls}, but found: {type(ep)}")
            return None
        except (ServiceUnavailableException, ConnectTimeout) as ex:
            logger.error(f"Failed to create endpoint with key={ek}: {ex}")
            return None

    def _get_endpoint_or_default(self, ek: str, default_ek: str, *, capability: str = "") -> LLMEndpoint | None:
        return (ek and self._get_endpoint(ek, capability)) or self._get_endpoint(default_ek, capability)

    def _get_response_from_image(self, payload: LLMPayload, resource_id, props: LLMCallProps) -> str:
        ek = props.endpoint_key

        dtype = self.file_storage.get_dtype(resource_id)
        mimetype: str | None = mimetypes.guess_type(resource_id)[0] or (dtype and IMAGE_MIME_TYPES[dtype])
        if not mimetype:
            logger.error(f"Failed to derive mimetype: {mimetype}")
            return ""

        sent: str = _parse_prompt_for_image(payload)
        ep = self._get_endpoint_or_default(ek, self.default_image_ek, capability="image")
        if ep is None:
            logger.error(f"Failed to get image endpoint for keys=('{ek}', '{self.default_image_ek}')")
            return ""

        file: bytes = self.file_storage.download(resource_id)
        get_image_response = getattr(ep, "get_image_response", None)
        if not get_image_response:
            logger.error(f"Not image endpoint: {ek}: {type(ep)}")
            return ""
        response_image: str = get_image_response(bytesimage=file, sentences=sent, mimetype=mimetype)
        logger.debug(f"Image response from {ep.__repr__()}: `{pretty_line(response_image)}`")
        return response_image

    def _upload_file(self, endpoint: LLMEndpoint, resource_id: ResourceId) -> str | None:
        # todo don't upload already loaded file
        # todo try-catch?
        resource_path = Path(resource_id)
        if not resource_path.exists():
            logger.error(f"Can not found resource_id={resource_id}")
            return None
        with resource_path.open("rb") as file_handle:
            upload_file = getattr(endpoint, "upload_file")
            if not upload_file:
                logger.warning(f"Not file endpoint: {endpoint}")
                return None
            uploaded_file = upload_file(file=file_handle)
            file_id = uploaded_file.id_
            logger.info(f"Uploaded file: {resource_id} -> {file_id}")
            return file_id

    def _get_response_from_payload(
        self, endpoint: LLMEndpoint, payload: LLMPayload, props: LLMCallProps = LCP
    ) -> LLMResponseExt:
        ek_pretty = f"(endpoint_key={props.endpoint_key})" if props.endpoint_key else ""
        retrier = retry_on_cond(
            title=f"#get_response{ek_pretty}", attempts=props.attempts, wait_seconds=self.wait_seconds
        )
        get_response = retrier(endpoint.get_response)
        payload_dict = {
            "messages": payload.model_dump()["messages"],
        }
        if payload.attachments:
            payload_dict["attachments"] = payload.attachments
        text = get_response(request=payload_dict) or ""
        return LLMResponseExt(text=text)

    def _get_response_ext(self, payload: LLMPayload, props: LLMCallProps) -> LLMResponseExt:
        ek = props.endpoint_key
        resource_id: str | None = payload.get_resource_id()

        if not resource_id:
            endpoint_b: LLMEndpoint | None = self._get_endpoint_or_default(ek, self.default_ek)
            if endpoint_b is None:
                logger.error(f"Failed to find endpoint: {ek}, default: {self.default_ek}")
                return RESPONSE_EMPTY
            return self._get_response_from_payload(endpoint_b, payload, props)

        dtype = self.file_storage.get_dtype(resource_id)

        if dtype in {"jpg", "png"}:
            text = self._get_response_from_image(payload, resource_id, props)
            return LLMResponseExt(text=text)

        if dtype in {"pdf", "csv", "txt"}:
            endpoint_f = self._get_endpoint_or_default(ek, self.default_file_ek, capability="file")
            if endpoint_f is None:
                logger.error(f"Failed to get file endpoint for keys=({ek}, {self.default_file_ek}")
                return RESPONSE_EMPTY

            # todo don't upload already loaded file
            # todo try-catch?
            file_id = self._upload_file(endpoint_f, resource_id)
            if file_id:
                payload_with_file = payload.with_attachments([[file_id]])
                logger.info(f"Sending request with file: {payload_with_file}")
                return self._get_response_from_payload(endpoint_f, payload_with_file, props)
            else:
                return self._get_response_from_payload(endpoint_f, payload, props)

        endpoint: LLMEndpoint | None = self._get_endpoint_or_default(ek, self.default_ek)
        if endpoint is None:
            logger.error(f"Failed to get file endpoint for keys=({ek}, {self.default_file_ek}")
            return RESPONSE_EMPTY

        return self._get_response_from_payload(endpoint, payload, props)

    # API

    def get_metadata(self) -> LLMHubMetadata:
        return self.config_llm.as_metadata()

    def get_response(self, *, request: LLMRequest, props: LLMCallProps = LCP) -> str:
        response_ext = self.get_response_ext(request=request, props=props)
        return response_ext.text

    def get_response_ext(self, *, request: LLMRequest, props: LLMCallProps = LCP) -> LLMResponseExt:
        start = time.time()
        payload: LLMPayload = LLMPayload.parse(request)
        response = self._get_response_ext(payload, props)
        elapsed = time.time() - start
        ek_pretty = f", endpoint_key={props.endpoint_key}" if props.endpoint_key else ""
        # todo fix: also show real time, without retries and waits
        logger.info(f"Ready in {elapsed:.2f} seconds ( {payload.show_pretty(detailed=True)}{ek_pretty} )")
        return response

    def get_embedding(self, *, prompt: str, props: LLMCallProps = LCP) -> list[float] | None:
        ek = props.endpoint_key
        endpoint: LLMEndpoint | None = self._get_endpoint_or_default(ek, self.default_ek)
        if endpoint is None:
            logger.error(f"Failed to get endpoint for keys=({ek}, {self.default_file_ek}")
            return None
        # todo move to library
        prompt_pretty = pretty_line(prompt)
        retrier = retry_on_cond(
            title=f"#get_embedding(endpoint_key={props.endpoint_key}), prompt={prompt_pretty}",
            attempts=props.attempts,
            condition=lambda embedding: any(map(abs, embedding)),
            wait_seconds=self.wait_seconds,
        )
        get_embedding = retrier(endpoint.get_embedding)
        return get_embedding(prompt)



from mmar_llm.airi_endpoint import AiriChatEndpoint
from mmar_llm.dummy_endpoint import DummyEndpoint
from mmar_llm.fusion_brain_endpoint import FusionBrainEndpoint
from mmar_llm.gigachat_endpoint import (
    GigaChatEndpoint,
    GigaChatSberdevicesEndpoint,
)
from mmar_llm.llm_endpoint import LLMEndpoint
from mmar_llm.llm_hub import LLMHub
from mmar_llm.llm_hub_config import LLMConfig, LLMEndpointConfig, LLMHubConfig
from mmar_llm.openrouter_endpoint import OpenRouterEndpoint
from mmar_llm.yandex_gpt_endpoint import YandexGPTEndpoint

__all__ = [
    "AiriChatEndpoint",
    "DummyEndpoint",
    "EndpointsConfig",
    "FusionBrainEndpoint",
    "GigaChatEndpoint",
    "GigaChatSberdevicesEndpoint",
    "LLMConfig",
    "LLMEndpoint",
    "LLMEndpointConfig",
    "LLMEndpointsConfig",
    "LLMHub",
    "LLMHubConfig",
    "OpenRouterEndpoint",
    "YandexGPTEndpoint",
    "create_endpoint",
]



import contextlib
import json

import requests

from mmar_mapi.api import LLMPayload, LLMRequest

from mmar_llm.llm_endpoint import LLMEndpoint
from mmar_llm.utils import dump_messages


class YandexGPTEndpoint(LLMEndpoint):
    def __init__(
        self,
        token: str,
        folder_id: str,
        iam_url: str = "https://iam.api.cloud.yandex.net/iam/v1/tokens",
        text_url: str = "https://llm.api.cloud.yandex.net/foundationModels/v1/completion",
        emb_url: str = "https://llm.api.cloud.yandex.net:443/foundationModels/v1/textEmbedding",
        warmup: bool = False,
    ):
        iam_token = requests.post(
            url=iam_url,
            data=json.dumps({"yandexPassportOauthToken": token}),
        ).json()["iamToken"]
        self.folder_id = folder_id
        self.text_url = text_url
        self.emb_url = (emb_url,)
        self.doc_uri = f"emb://{self.folder_id}/text-search-doc/latest"
        self.query_uri = f"emb://{self.folder_id}/text-search-query/latest"
        self.headers = {
            "Authorization": f"Bearer {iam_token}",
            "Content-Type": "application/json",
        }
        self._DIM: int = 256
        self._ZEROS: list[float] = [0 for _ in range(self._DIM)]
        self._ERROR_MESSAGE: str = ""
        if warmup:
            self.warmup()

    def __call__(self) -> None:
        return None

    def get_response(self, *, request: LLMRequest) -> str:
        payload = LLMPayload.parse(request)
        messages_json = dump_messages(payload)
        data = {
            "model_uri": f"gpt://{self.folder_id}/yandexgpt/latest",
            "messages": messages_json,
        }
        response = requests.post(url=self.text_url, headers=self.headers, json=data).json()
        return response["result"]["alternatives"][0]["message"]["text"]

    def get_embedding_custom(self, sentence: str, input_is_long: bool = False) -> list[float]:
        data = {
            "modelUri": self.doc_uri if input_is_long else self.query_uri,
            "text": sentence,
        }
        with contextlib.suppress(Exception):
            response = requests.post(url=self.emb_url, headers=self.headers, json=data).json()
            return response["embedding"]
        return self._ZEROS

    def get_embeddings(self, sentences: list[str], input_is_long: bool = False) -> list[list[float]]:
        return [self.get_embedding(sentence, input_is_long) for sentence in sentences]



from loguru import logger

from mmar_llm.models import UnsupportedModelException
from mmar_llm.openrouter_endpoint import OpenRouterEndpoint


class AiriChatEndpoint(OpenRouterEndpoint):
    def __init__(
        self,
        model_id: str,
        base_url: str,
        api_key: str = "",
        emb_dim: int = 1024,
        verify: bool = True,
    ) -> None:
        super().__init__(model_id=model_id, base_url=base_url, api_key=api_key, emb_dim=emb_dim, verify=False)
        self.api_key = api_key
        self.verify = verify
        # todo fix: move in warmup?
        # self.check_current(model_id=model_id)

    def check_current(self, model_id: str) -> None:
        model_list = self.get_available_models()

        if model_id not in model_list:
            logger.error("Model '%s' not found. Available models: %s", model_id, model_list)
            raise UnsupportedModelException(model_id, model_list)
        else:
            logger.info("Model '%s' is found.", model_id)

    def get_available_models(self) -> list[str]:
        models_data = self._model.models.list().data
        models_ids = [model.id for model in models_data]
        logger.info(f"Available models: {models_ids}")
        return models_ids



import importlib
import tiktoken

from mmar_mapi.api import LLMPayload


def count_tokens(sentences: list[str]) -> list[int]:
    encoding = tiktoken.get_encoding("cl100k_base")
    return [len(encoding.encode(sentence)) for sentence in sentences]


# todo specify better type
def dump_messages(payload: LLMPayload) -> list[dict]:
    messages_json = [{"role": msg.role, "content": msg.content} for msg in payload.messages]
    return messages_json


def load_dynamically(object_path: str):
    module_path, class_name = object_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)



from collections import defaultdict
from types import SimpleNamespace
from typing import Annotated, NotRequired, Protocol

from mmar_mapi.services import LLMEndpointMetadata, LLMHubMetadata
from pydantic import AfterValidator, BaseModel

StrDict = dict[str, str | bool | int]


class LLMEndpointConfig(BaseModel):
    key: str
    class_path: str
    caption: str
    args: StrDict
    extra: StrDict | None = None

    def as_metadata(self) -> LLMEndpointMetadata:
        return LLMEndpointMetadata(key=self.key, caption=self.caption)


def _validate_unique_keys(endpoints: list[LLMEndpointConfig]) -> list[LLMEndpointConfig]:
    endpoints_by_keys = defaultdict(list)
    for epc in endpoints:
        endpoints_by_keys[epc.key].append(epc)
    errors = []
    for key, endpoints_with_key in endpoints_by_keys.items():
        if len(endpoints_with_key) > 1:
            errors.append(f"For key={key} found many endpoints: {endpoints_with_key}")
    if errors:
        raise ValueError("\n".join(errors))
    return endpoints


LLMEndpointConfigs = Annotated[list[LLMEndpointConfig], AfterValidator(_validate_unique_keys)]


class LLMConfig(BaseModel):
    endpoints: LLMEndpointConfigs

    default_endpoint_key: str = ""
    default_image_endpoint_key: str = ""
    default_file_endpoint_key: str = ""

    warmup: bool = False
    wait_seconds_on_llm_retry: float = 1.0

    def as_metadata(self) -> LLMHubMetadata:
        # we can do `return LLMHubMetadata(**self.model_dump())`, but better to pass manually to ensure that no private data passed
        return LLMHubMetadata(
            endpoints=[ep.as_metadata() for ep in self.endpoints],
            default_endpoint_key=self.default_endpoint_key,
        )

    def get_endpoint_config(self, endpoint_key: str) -> LLMEndpointConfig | None:
        return next((epc for epc in self.endpoints if epc.key == endpoint_key), None)


class LLMHubConfig(Protocol):
    llm: LLMConfig
    files_dir: str | None
    validate_endpoints: NotRequired[bool] = True


LLM_CONFIG_EMPTY = LLMConfig(endpoints=[])
LLM_HUB_CONFIG_EMPTY = SimpleNamespace(llm=LLM_CONFIG_EMPTY, files_dir=None)



import re
import warnings
from base64 import b64decode, b64encode
from functools import partial
from typing import Literal

from gigachat import GigaChat
from gigachat._types import FileTypes
from gigachat.models import UploadedFile
from mmar_mapi.services import LLMPayload, LLMRequest, LLMResponseExt

from mmar_llm.llm_endpoint import LLMEndpoint
from mmar_llm.utils import dump_messages

PATTERN_CID = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
CPATTERN_CID = re.compile(PATTERN_CID)
SCOPES = {"GIGACHAT_API_PERS", "GIGACHAT_API_B2B", "GIGACHAT_API_CORP"}
SCOPE_DEFAULT = "GIGACHAT_API_CORP"
BASE_URL_DEFAULT = "https://gigachat.devices.sberbank.ru/api/v1"
MODEL_DEFAULT = "GigaChat-2-Max"


def _validate_cid(field_value, field: str):
    if CPATTERN_CID.fullmatch(field_value):
        return
    warnings.warn(f"Maybe '{field}' is invalid: not matched by '{PATTERN_CID}'")


_validate_client_id = partial(_validate_cid, field="client_id")
_validate_client_secret = partial(_validate_cid, field="client_secret")


def _validate_scope(scope):
    if scope in SCOPES:
        return
    warnings.warn(f"Maybe '{scope}' is invalid: not in {SCOPES}")


def _make_kwargs_auth(
    user: str | None,
    password: str | None,
    client_id: str | None,
    client_secret: str | None,
    authorization_key: str | None,
):
    if user and password:
        return dict(user=user, password=password)
    if client_id and client_secret:
        _validate_client_id(client_id)
        _validate_client_secret(client_secret)
        creds = b64encode(f"{client_id}:{client_secret}".encode("utf-8")).decode("utf-8")
        return dict(credentials=creds)
    if authorization_key:
        authorization_key_decode = b64decode(authorization_key).decode()
        if ":" not in authorization_key_decode:
            warnings.warn("Maybe bad value for 'authorization_key': expected `:`")
        else:
            client_id, client_secret = authorization_key_decode.split(":", 1)
            _validate_client_id(client_id)
            _validate_client_secret(client_secret)
        return dict(credentials=authorization_key)
    raise ValueError(f"Unexpected combination of auth args: {user=}, {password=}, {client_id=}, {client_secret=}")


class GigaChatEndpoint(LLMEndpoint):
    def __init__(
        self,
        user: str | None = None,
        password: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        authorization_key: str | None = None,
        scope: str = SCOPE_DEFAULT,
        model_id: str = MODEL_DEFAULT,
        base_url: str = BASE_URL_DEFAULT,
        temperature: float = 0.0,
        timeout: float | None = None,
    ) -> None:
        _validate_scope(scope)
        kwargs_pre = dict(
            base_url=base_url,
            scope=scope,
            model=model_id,
            verify_ssl_certs=False,
            profanity_check=False,
            timeout=timeout,
        )
        kwargs_auth = _make_kwargs_auth(user, password, client_id, client_secret, authorization_key)
        kwargs = kwargs_pre | kwargs_auth

        self._model = GigaChat(**kwargs)
        self._model_id = model_id
        self._base_url = base_url
        self.temperature = temperature
        # todo fix: eliminate, move to LLMHub
        self._dim = 1024

    def get_response_ext(self, *, request: LLMRequest) -> LLMResponseExt:
        payload = LLMPayload.parse(request)
        messages_json = dump_messages(payload)
        payload_json = {"messages": messages_json, "temperature": self.temperature}

        # todo ensure that content is ok
        text = self._model.chat(payload_json).choices[0].message.content
        return LLMResponseExt(text=text)

    def get_embedding(self, *, prompt: str) -> list[float]:
        return self._model.embeddings([prompt]).data[0].embedding

    def upload_file(
        self,
        file: FileTypes,
        purpose: Literal["general", "assistant"] = "general",
    ) -> UploadedFile:
        return self._model.upload_file(file, purpose)

    def __repr__(self):
        class_name = type(self).__name__
        url_info = "" if "https://gigachat.devices.sberbank.ru/api/v1" == self._base_url else f", url={self._base_url}"
        model_info = f"model_id={self._model_id}"
        return f"{class_name}({model_info}{url_info})"

    def __str__(self):
        return repr(self)


class GigaChatSberdevicesEndpoint(GigaChatEndpoint):
    def __init__(self, **kwargs):
        if "base_url" in kwargs:
            base_url = kwargs["base_url"]
            raise ValueError(f"For key 'base_url' found value {base_url}, expected nothing")
        super().__init__(base_url="https://gigachat.sberdevices.ru/v1", **kwargs)



from mmar_mapi.services import LLMRequest, LLMResponseExt


class LLMEndpoint:
    def get_response_ext(self, *, request: LLMRequest) -> LLMResponseExt:
        raise NotImplementedError

    def get_embedding(self, *, prompt: str) -> list[float]:
        raise NotImplementedError

    # helpers

    def get_response(self, *, request: LLMRequest) -> str:
        response_ext = self.get_response_ext(request=request)
        return response_ext.text


class LLMEndpointFacade(LLMEndpoint):
    def __init__(self, base: LLMEndpoint, decorator):
        self._base = base

        self._get_response = decorator(self._base.get_response)
        self._get_response_ext = decorator(self._base.get_response_ext)
        self._get_embedding = decorator(self._base.get_embedding)

    def get_response(self, *, request: LLMRequest) -> str:
        return self._get_response(request=request)

    def get_response_ext(self, *, request: LLMRequest) -> LLMResponseExt:
        return self._get_response_ext(request=request)

    def get_embedding(self, *, prompt: str) -> list[float]:
        return self._get_embedding(prompt=prompt)



import json
import time
from enum import StrEnum, auto

import requests
from mmar_mapi.api import LLMRequest
from openai.types.chat import ChatCompletionMessageParam

from mmar_llm.llm_endpoint import LLMEndpoint


class ResponseStatus(StrEnum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    INITIAL = auto()
    DONE = auto()
    PROCESSING = auto()
    FAIL = auto()


class FusionBrainEndpoint(LLMEndpoint):
    def __init__(
        self, api_key: str, secret_key: str, warmup: bool = True, max_gen_attempts: int = 6, gen_ping_timeout: int = 5
    ):
        self.auth_headers = {
            "X-Key": f"Key {api_key}",
            "X-Secret": f"Secret {secret_key}",
        }
        self.base_url = "https://api-key.fusionbrain.ai"
        self.pipeline_id: str = self._get_pipeline()

        self.max_gen_attempts: int = max_gen_attempts
        self.gen_ping_timeout: int = gen_ping_timeout

        if warmup:
            self.warmup()

    def get_response(self, *, request: LLMRequest) -> str:
        raise NotImplementedError

    def get_response_custom(
        self, sentence: str, width: int = 1024, height: int = 1024, negative_prompt: str = "", style: str = ""
    ) -> str:
        uuid = self._start_generation(
            sentence, width=width, height=height, negative_prompt=negative_prompt, style=style
        )
        response = self._end_generation(uuid)
        if result := response.get("result"):
            return result["files"][0]

        if response["status"] in [ResponseStatus.INITIAL, ResponseStatus.PROCESSING]:
            raise RuntimeError(
                f"Max generation time of {self.max_gen_attempts * self.gen_ping_timeout} seconds exceeded!"
            )

        raise RuntimeError("Error getting generation result " + response["status"])

    def _get_pipeline(self) -> str:
        response = requests.get(f"{self.base_url}/key/api/v1/pipelines", headers=self.auth_headers)
        if response.status_code >= 400:
            raise RuntimeError(f"Error getting pipeline id {response.status_code} {response.reason}")

        data = response.json()
        return data[0]["id"]

    def _start_generation(
        self, sentence: str, width: int = 1024, height: int = 1024, negative_prompt: str = "", style: str = ""
    ) -> str:
        params = {
            "type": "GENERATE",
            "numImages": 1,
            "width": width,
            "height": height,
            "style": style,
            "negativePromptDecoder": negative_prompt,
            "generateParams": {"query": f"{sentence}"},
        }

        data = {"pipeline_id": (None, self.pipeline_id), "params": (None, json.dumps(params), "application/json")}
        response = requests.post(f"{self.base_url}/key/api/v1/pipeline/run", headers=self.auth_headers, files=data)
        if response.status_code >= 400:
            raise RuntimeError(
                f"Error sending generation request to FusionBrain {response.status_code} {response.reason}"
            )

        return response.json()["uuid"]

    def _check_generation(self, uuid: str) -> dict:
        response = requests.get(f"{self.base_url}/key/api/v1/pipeline/status/{uuid}", headers=self.auth_headers)
        return response.json()

    def _end_generation(self, uuid: str) -> dict:
        attempts = 0
        while attempts < self.max_gen_attempts:
            data = self._check_generation(uuid)
            if data["status"] == ResponseStatus.DONE:
                return data

            attempts += 1
            time.sleep(self.gen_ping_timeout)

        return data

    def warmup(self) -> None:
        pass

    def get_response_by_payload(self, payload: list[ChatCompletionMessageParam]) -> str:
        raise NotImplementedError()

    def get_embedding(self, *, prompt: str) -> list[float]:
        raise NotImplementedError()
